# TODO: (FEAT) support additional apache formats (http://ossec-docs.readthedocs.org/en/latest/log_samples/apache/apache.html)  # NOQA
# TODO: (FEAT) maybe consolidate Json and Custom formatters to the same class, to let other custom formatters use them both for each type implemented  # NOQA
# TODO: (FEAT) implement Tomcat Formatter

import random
from abc import abstractmethod, ABCMeta
from faker import Factory
from format_mappings import InHouseFaker
import json
import uuid


DEFAULT_CUSTOM_FORMAT = [
    'current_date_time', ' ', 'uuid', ' ', 'level', ': ',
    'module', ' - ', 'free_email'
]
DEFAULT_CUSTOM_DATA = {
    'current_date_time': '$RAND',
    'uuid': [str(uuid.uuid1()) for i in xrange(100)],
    'level': ['WARNING', 'ERROR', 'DEBUG', 'INFO', 'CRITICAL'],
    'module': ['module1', 'module2'],
    'free_email': '$RAND',
}
DEFAULT_JSONIFY = True


def fake_data(data_type):
    """returns fake data for the data type requested.

    will try and get fake data from the local format mappings.
    if fake data for the specific data type wasn't found, it will try and use
    fake-factory to fake the data.

    :param string data_type: the type of data to fake.
    :rtype: string
    """
    fake = Factory.create()
    try:
        return getattr(InHouseFaker(), 'default')(data_type)
    except KeyError:
        if hasattr(InHouseFaker, data_type):
            return getattr(InHouseFaker(), data_type)()
        elif hasattr(fake, data_type):
            return str(getattr(fake, data_type)())
        # except AttributeError:
        print('cannot randomize data for {0}. run "feedr list fake" '
              'to print a list of possible types.'.format(data_type))
        raise RuntimeError('cannot randomize data type {0}'.format(
            data_type))


class BaseFormatter(object):
    """base class for all formatters
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def __init__(self, config):
        return

    @abstractmethod
    def generate_data(self):
        return

    def f(self, config, name):
        """retrieves configuration keys

        if config wasn't supplied or a key doesn't exist, returns $RAND
        which will initiate data faking
        """
        return config['data'].get(name, '$RAND') if config else '$RAND'


class CustomFormatter(BaseFormatter):
    """returns a generated log string in a custom format

    this is also a formatter other formatters
    can rely on to generate application specific logs.
    see the ApacheAccessFormatter class for reference.

    for every item in the format list, if an item in the data dict
    corresponds with it and the field's data equals "$RAND", use faker
    to fake an item for it. else, choose one item from the list randomly.
    if there no item in the data to correspond with the format, it will
    just append to format's field name to the log.

    example:

    .. code-block:: python

     'CustomFormatter': {
         'format': ['name', ' - ', 'level'],
         'data': {
             'name': $RAND,
             'level': ['ERROR', 'DEBUG', 'INFO', 'CRITICAL'],
         }
     }

    the output of the above example might be:

    .. code-block:: python

     Sally Fields - ERROR
     or
     Jason Banks - DEBUG
     or
     Danny Milwee - ERROR
     or
     ...
    """
    def __init__(self, config):
        self.format = config.get('format', DEFAULT_CUSTOM_FORMAT)
        self.data = config.get('data', DEFAULT_CUSTOM_DATA)

    def generate_data(self):
        """this will generate a message according to `self.format`
        with data from `self.data`.

        all fields in the data dict will be iterated over and matched to
        the items in the format list. if a match is found and $RAND is set
        in one of the fields, random data will be generated for that field.
        If not, data will be chosen from the list.
        If no match is found, the explicit item in the format list will be
        appended.

        example:

        .. code-block:: python

         format = ['Mr. ' 'first_name', 'last_name']
         data = {
             'first_name': ['Jason, Josh]',
             'last_name': '$RAND'
         }

        the output of the above example might be:

        .. code-block:: python

         'Mr. Jason Williams'
         or
         'Mr. Josh Brolin'
         or
         'Mr. Jason Bananas'
         ...
        """
        log = ''
        # iterate over the format
        for field_name in self.format:
            # for each field in the data dictionary
            for field, data in self.data.items():
                # if the field name exists in the format and the data
                if field_name == field:
                    # and rand is set
                    if data == '$RAND':
                        # fake the data
                        log += fake_data(field_name)
                    else:
                        # else choose randomly from the field
                        log += random.choice(self.data[field_name])
            # if the field doesn't exist in the data, only in the format
            if field_name not in self.data.keys():
                # we'll assume that the field name in the format itself
                # should be appended to the log.
                log += field_name
        return log


class JsonFormatter(BaseFormatter):
    """generates log strings in json format (or leave as dict)

    all fields in the data dict will be iterated over.
    if $RAND is set in one of the fields, random data will be generated
    for that field. If not, data will be chosen from the list.

    example:

    .. code-block:: python

     'JsonFormatter': {
         'data': {
             'date_time': '$RAND',
             'level': ['ERROR', 'DEBUG'],
             'address': '$RAND',
         }
     },

    the output of the above example might be:

    .. code-block:: python

     {"date_time": "2006-11-05 13:31:09", "name": "Miss Nona Breitenberg DVM", "level": "ERROR"}  # NOQA
     or
     {"date_time": "1985-01-20 11:41:16", "name": "Almeda Lindgren", "level": "DEBUG"}  # NOQA
     or
     {"date_time": "1973-05-21 01:06:04", "name": "Jase Heaney", "level": "DEBUG"}  # NOQA
     or
     ...
    """
    def __init__(self, config):
        self.data = config.get('data', DEFAULT_CUSTOM_DATA)
        self.jsonify = config.get('jsonify', DEFAULT_JSONIFY)
        self.stringify = config.get('stringify', False)

    def generate_data(self):
        # TODO: (FEAT) support randomizing data fields in Json formatter
        log = {}
        for field, data in self.data.items():
            if data == '$RAND':
                log[field] = fake_data(field)
            else:
                log[field] = random.choice(data)
        if self.jsonify:
            log = json.dumps(log)
        if self.stringify:
            log = str(log)
        return log


class ApacheAccessFormatter(CustomFormatter):
    """returns an apache-access-log like string

    you can easily construct new formatters by inheriting the custom formatter.

    all you have to do is specify the format and the data. a helper
    method `f` is supplied in the `BaseFormatter` Class that will allow you to
    retrieve basic formatter configuration for your fields.

    """
    # 192.168.72.177 - - [22/Dec/2002:23:32:19 -0400] "GET /search.php HTTP/1.1" 400 1997 www.yahoo.com "-" "Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; ...)" "-"  # NOQA
    # %{IPORHOST:clientip} %{USER:ident} %{USER:auth} \[%{HTTPDATE:timestamp}\] "(?:%{WORD:verb} %{NOTSPACE:request}(?: HTTP/%{NUMBER:httpversion})?|%{DATA:rawrequest})" %{NUMBER:response} (?:%{NUMBER:bytes}|-)  # NOQA
    def __init__(self, config):
        self.format = [
            'ipv4', ' - - [', 'current_day_of_month', '/',
            'current_month_name_short', '/',
            'current_year', ':', 'current_time', ' ',
            'current_time_zone_number', '] "', 'http_verbs',
            ' /', 'uri_path', ' ', 'http_versions', '" ',
            'http_error_codes', ' ', 'random_int'
        ]
        self.data = {
            'ipv4': self.f(config, 'ipv4'),
            'current_day_of_month': self.f(config, 'current_day_of_month'),
            'current_month_name_short': self.f(config, 'current_month_name_short'),  # NOQA
            'current_year': self.f(config, 'current_year'),
            'current_time': self.f(config, 'current_time'),
            'current_time_zone_number': self.f(config, 'current_time_zone_number'),  # NOQA
            'http_versions': self.f(config, 'http_versions'),
            'http_verbs': self.f(config, 'http_verbs'),
            'uri_path': self.f(config, 'uri_path'),
            'http_error_codes': self.f(config, 'http_error_codes'),
            'random_int': self.f(config, 'random_int'),
        }


class ApacheAccessExFormatter(CustomFormatter):
    """returns an apache-extended-access-log like string"""
    # http://httpd.apache.org/docs/2.2/logs.html
    # 127.0.0.1 - frank [10/Oct/2000:13:55:36 -0700] "GET /apache_pb.gif HTTP/1.0" 200 2326  # NOQA
    # 192.168.2.20 - - [28/Jul/2006:10:27:10 -0300] "GET /cgi-bin/try/ HTTP/1.0" 200 3395  # NOQA
    # %{IPORHOST:clientip} %{USER:ident} %{USER:auth} \[%{HTTPDATE:timestamp}\] "(?:%{WORD:verb} %{NOTSPACE:request}(?: HTTP/%{NUMBER:httpversion})?|%{DATA:rawrequest})" %{NUMBER:response} (?:%{NUMBER:bytes}|-)  # NOQA
    def __init__(self, config):
        self.format = [
            'ipv4', ' - - [', 'current_day_of_month', '/',
            'current_month_name_short', '/',
            'current_year', ':', 'current_time', ' ',
            'current_time_zone_number', '] "', 'http_verbs',
            ' /', 'uri_path', ' ', 'http_versions', '" ',
            'http_error_codes', ' ', 'random_int', ' "',
            'uri', '" "', 'user_agent', '"'
        ]
        self.data = {
            'ipv4': self.f(config, 'ipv4'),
            'current_day_of_month': self.f(config, 'current_day_of_month'),
            'current_month_name_short': self.f(config, 'current_month_name_short'),  # NOQA
            'current_year': self.f(config, 'current_year'),
            'current_time': self.f(config, 'current_time'),
            'current_time_zone_number': self.f(config, 'current_time_zone_number'),  # NOQA
            'http_versions': self.f(config, 'http_versions'),
            'http_verbs': self.f(config, 'http_verbs'),
            'uri_path': self.f(config, 'uri_path'),
            'http_error_codes': self.f(config, 'http_error_codes'),
            'random_int': self.f(config, 'random_int'),
            'uri': self.f(config, 'uri'),
            'user_agent': self.f(config, 'user_agent'),
        }


class ApacheErrorFormatter(CustomFormatter):
    """returns an apache-error-log like string"""
    # [Fri Dec 16 01:46:23 2005] [error] [client 1.2.3.4] Directory index forbidden by rule: /home/test/  # NOQA
    # [Mon Dec 19 23:02:01 2005] [error] [client 1.2.3.4] user test: authentication failure for "/~dcid/test1": Password Mismatch  # NOQA
    def __init__(self, config):
        self.format = [
            '[', 'current_day_of_week_short', ' ', 'current_month_name_short',
            ' ',
            'current_day_of_month', ' ', 'current_time', ' ', 'current_year',
            '] [',
            'syslog_error_levels_lower', '] [client ', 'ipv4', '] ',
            'catch_phrase'
        ]
        self.data = {
            'current_day_of_week_short': self.f(config, 'current_day_of_week_short'),  # NOQA
            'ipv4': self.f(config, 'ipv4'),
            'current_day_of_month': self.f(config, 'current_day_of_month'),
            'current_month_name_short': self.f(config, 'current_month_name_short'),  # NOQA
            'current_year': self.f(config, 'current_year'),
            'current_time': self.f(config, 'current_time'),
            'catch_phrase': self.f(config, 'catch_phrase'),
            'syslog_error_levels_lower': self.f(config, 'syslog_error_levels_lower'),  # NOQA
        }



#!/usr/bin/env python

# TODO: (FEAT) add graphite - https://github.com/daniellawrence/graphitesend
# TODO: https://github.com/globocom/tornado-es

import pika
import logging
import os
from abc import abstractmethod, ABCMeta
from datetime import date
from time import sleep

# TRANSPORT SPECIFIC
from logentries import LogentriesHandler
import pymongo
from elasticsearch import Elasticsearch
from influxdb import client as influxdb
import urllib2

# import pretty table for statistical data visualization
from prettytable import PrettyTable

# FILE DEFAULTS
DEFAULT_FILE_PATH = 'generated.log'
DEFAULT_MAXBYTES = 10000000
DEFAULT_BACKUPS = 20

# AMQP DEFAULTS
# queue name for packager events
DEFAULT_AMQP_QUEUE = ''
# broker exchange
DEFAULT_AMQP_EXCHANGE = ''
DEFAULT_AMQP_DELIVERY_MODE = 2

# ES DEFAULTS
DEFAULT_ES_PORT = '9200'
DEFAULT_ES_URL_PREFIX = ''
DEFAULT_ES_TIMEOUT = '10'

d = date.today()
DEFAULT_ES_INDEX = 'logstash-{0}.{1}.{2}'.format(d.year, d.month, d.day)
DEFAULT_ES_DOC_TYPE = 'doc'
DEFAULT_ES_SLEEP = 3

# LOGGLY DEFAULTS
DEFAULT_LOGGLY_DOMAIN = 'logs-01.loggly.com'

# MONGO DEFAULTS
DEFAULT_MONGO_PORT = 27017
DEFAULT_MONGO_DB = 'test'
DEFAULT_MONGO_COLLECTION = 'my_collection'
DEFAULT_MONGO_SLEEP = 1

# INFLUX DEFAULS
DEFAULT_INFLUX_PORT = 8086
DEFAULT_INFLUX_DB = 'metrics'
DEFAULT_INFLUX_USER = 'root'
DEFAULT_INFLUX_PASSWORD = 'root'


class BaseTransport(object):

    __metaclass__ = ABCMeta

    @abstractmethod
    def __init__(self, config):
        return

    @abstractmethod
    def configure(self):
        return

    @abstractmethod
    def send(self, client, log):
        return


class FileTransport(BaseTransport):
    """a RotatingFileHandler transport implementation"""
    def __init__(self, config):
        self.file_path = config.get('file', DEFAULT_FILE_PATH)
        self.max_bytes = config.get('max_bytes', DEFAULT_MAXBYTES)
        self.backups_count = config.get('backups_count', DEFAULT_BACKUPS)

    def configure(self):
        if os.path.exists(self.file_path):
            os.remove(self.file_path)
        logger = logging.getLogger('feedr')
        handler = logging.handlers.RotatingFileHandler(
            self.file_path, maxBytes=self.max_bytes,
            backupCount=self.backups_count)
        logger.addHandler(handler)
        logger.setLevel(logging.DEBUG)
        return logger

    def send(self, client, data):
        for piece in data:
            client.info(piece)

    def close(self):
        pass

    def get_data(self):
        # https://code.google.com/p/prettytable/wiki/Tutorial
        # you can also get data from csv's, htmls, etc...
        # so for instance, you could query elasticsearch right here,
        # get the data, and write it here, so that you can make sure
        # all logs you wrote actually made it through.
        data_table = PrettyTable(["Type", "Value"])
        # make sure the "type" field is always aligned to the left.
        data_table.align["Type"] = "l"
        # ...
        data_table.padding_width = 1
        with open(self.file_path) as f:
            # populate the table
            # TODO: (BUG) fix bug where rotating files will return a bad line
            # TODO: (BUG) count if more than one file is generated
            data_table.add_row(["lines written", sum(1 for _ in f)])
        return data_table


class AMQPTransport(BaseTransport):
    """an amqp transport implementation"""
    def __init__(self, config):
        try:
            self.host = config['host']
        except KeyError as ex:
            raise RuntimeError('configuration incomplete: {0}'.format(ex))
        self.queue = config.get('queue', DEFAULT_AMQP_QUEUE)
        self.exchange = config.get('exchange', DEFAULT_AMQP_EXCHANGE)
        self.exchange_type = config.get('exchange_type', 'fanout')
        self.routing_key = config.get('routing_key', self.queue)
        self.delivery_mode = config.get('deliver_mode',
                                        DEFAULT_AMQP_DELIVERY_MODE)
        self.auto_delete = config.get('auto_delete', True)
        self.durable = config.get('durable', True)
        self.exclusive = config.get('exclusive', False)

    def configure(self):
        if not self.host:
            raise RuntimeError('no host defined')
        try:
            self.connection = pika.BlockingConnection(
                pika.ConnectionParameters(self.host))
        except:
            raise RuntimeError('could not connect to host')

        settings = {
            'auto_delete': self.auto_delete,
            'durable': self.durable,
            'exclusive': self.exclusive
        }
        client = self.connection.channel()
        if len(self.exchange) > 0:
            client.exchange_declare(
                exchange=self.exchange, type=self.exchange_type)
        client.queue_declare(queue=self.queue, **settings)
        return client

    def send(self, client, data):
        for piece in data:
            client.basic_publish(exchange=self.exchange,
                                 routing_key=self.routing_key,
                                 body=piece,
                                 properties=pika.BasicProperties(
                                     delivery_mode=self.delivery_mode))

    # TODO: (FEAT) support connection closing
    def close(self):
        self.connection.close()


class UDPTransport(BaseTransport):
    """a udp transport implementation"""
    def __init__(self, config):
        try:
            self.host = config['host']
            self.port = config['port']
        except KeyError as ex:
            raise RuntimeError('configuration not complete: {0}'.format(ex))

    def configure(self):
        logger = logging.getLogger('feedr')
        handler = logging.handlers.DatagramHandler(self.host, self.port)
        logger.addHandler(handler)
        logger.setLevel(logging.DEBUG)
        return logger

    def send(self, client, data):
        for piece in data:
            client.debug(piece)

    def close(self):
        pass


class StreamTransport(BaseTransport):
    """a shell stream transport implementation"""
    def __init__(self, config):
        pass

    def configure(self):
        logger = logging.getLogger('feedr')
        handler = logging.StreamHandler()
        logger.addHandler(handler)
        logger.setLevel(logging.DEBUG)
        return logger

    def send(self, client, data):
        for piece in data:
            client.debug(piece)


class ElasticsearchTransport(BaseTransport):
    """an Elasticsearch transport implementation"""
    def __init__(self, config):
        try:
            self.host = config.get('host')
            self.port = config.get('port', DEFAULT_ES_PORT)
            self.url_prefix = config.get('url_prefix', DEFAULT_ES_URL_PREFIX)
            self.timeout = config.get('timeout', DEFAULT_ES_TIMEOUT)

            self.index = config.get('index', DEFAULT_ES_INDEX)
            self.doc_type = config.get('doc_type', DEFAULT_ES_DOC_TYPE)

            self.sleep = config.get('sleep', DEFAULT_ES_SLEEP)
        except KeyError as ex:
            raise RuntimeError('configuration not complete: {0}'.format(ex))

    def configure(self):
        es = Elasticsearch(
            host=self.host,
            port=int(self.port),
            url_prefix=self.url_prefix,
            timeout=float(self.timeout)
        )
        es.indices.create(index=self.index, ignore=400)
        self.indices_client = es.indices
        self.docs = self.indices_client.stats(
            index=self.index)['_all']['total']['docs']['count']
        return es

    def send(self, client, data):
        for piece in data:
            client.index(
                index=self.index,
                doc_type=self.doc_type,
                body=piece
            )

    def close(self):
        pass

    def get_data(self):
        sleep(self.sleep)
        data_table = PrettyTable(["Type", "Value"])
        data_table.align["Type"] = "l"
        data_table.padding_width = 1
        self.current_docs = self.indices_client.stats(
            index=self.index)['_all']['total']['docs']['count']
        data_table.add_row(["docs before", self.docs])
        data_table.add_row(["docs after", self.current_docs])
        data_table.add_row(["docs written", self.current_docs - self.docs])
        return data_table


class LogentriesTransport(BaseTransport):
    """a logentries transport implementation"""
    def __init__(self, config):
        try:
            self.token = config['token']
        except KeyError as ex:
            raise RuntimeError('configuration not complete: {0}'.format(ex))

    def configure(self):
        logger = logging.getLogger('logentries')
        handler = LogentriesHandler(self.token)
        logger.addHandler(handler)
        logger.setLevel(logging.DEBUG)
        return logger

    def send(self, client, data):
        for piece in data:
            client.debug(piece)

    def close(self):
        pass


class LogglyTransport(BaseTransport):
    """a Loggly transport implementation"""
    # TODO: (IMPRV) check out https://github.com/kennedyj/loggly-handler/
    def __init__(self, config):
        try:
            self.domain = config.get('url', DEFAULT_LOGGLY_DOMAIN)
            self.token = config['token']
        except KeyError as ex:
            raise RuntimeError('configuration not complete: {0}'.format(ex))

    def configure(self):
        logger = "http://{0}/inputs/{1}/tag/python/".format(
            self.domain, self.token)
        return logger

    def send(self, client, data):
        for piece in data:
            piece = "PLAINTEXT=" + urllib2.quote(piece)
        urllib2.urlopen(client, piece)

    def close(self):
        pass


class MongoDBTransport(BaseTransport):
    """a MongoDB transport implementation"""
    def __init__(self, config):
        try:
            self.host = config['host']
            self.port = config.get('port', DEFAULT_MONGO_PORT)
            self.db = config.get('db', DEFAULT_MONGO_DB)
            self.collection = config.get(
                'collection', DEFAULT_MONGO_COLLECTION)
            self.sleep = config.get('sleep', DEFAULT_MONGO_SLEEP)
        except KeyError as ex:
            raise RuntimeError('configuration not complete: {0}'.format(ex))

    def configure(self):
        mongo = pymongo.MongoClient(self.host, self.port)
        self.collection_client = mongo[self.db][self.collection]
        self.docs = self.collection_client.count()
        return self.collection_client

    def send(self, client, data):
        for piece in data:
            client.save(piece)

    def close(self):
        pass

    def get_data(self):
        sleep(self.sleep)
        data_table = PrettyTable(["Type", "Value"])
        data_table.align["Type"] = "l"
        data_table.padding_width = 1
        current_docs = self.collection_client.count()
        data_table.add_row(["docs before", self.docs])
        data_table.add_row(["docs after", current_docs])
        data_table.add_row(["docs written", current_docs - self.docs])
        return data_table


class InfluxDBTransport(BaseTransport):
    """an InfluxDB transport implementation"""
    def __init__(self, config):
        try:
            self.host = config['host']
            self.port = config.get('port', DEFAULT_INFLUX_PORT)
            self.username = config.get('username', DEFAULT_INFLUX_USER)
            self.password = config.get('password', DEFAULT_INFLUX_PASSWORD)
            self.database = config.get('database', DEFAULT_INFLUX_DB)
        except KeyError as ex:
            raise RuntimeError('configuration not complete: {0}'.format(
                ex.message))

    def configure(self):
        self.db = influxdb.InfluxDBClient(
            self.host, self.port, self.username, self.password, self.database)
        try:
            self.db.create_database(self.database)
        # TODO: (IMPRV) handle specific exception
        except:
            pass
        return self.db

    def send(self, client, data):
        for piece in data:
            client.write_points([piece])

    def close(self):
        pass

    def get_data(self):
        pass



# TODO: (FEAT) support for different statistical distributions
# TODO: (FEAT) support throughput testing thru http://linux.die.net/man/1/pv
# TODO: (IMPRV) remove six and dictconfig from coverage report


import logging
import logger
import os
import sys
from time import sleep
import datetime
import random

# import transports and formatters
import transports as trans
import formatters as forms

# import fakers
from faker import Factory
import format_mappings as fm


DEFAULT_BASE_LOGGING_LEVEL = logging.INFO
DEFAULT_VERBOSE_LOGGING_LEVEL = logging.DEBUG

DEFAULT_CONFIG_FILE = 'config.py'
DEFAULT_GAP = 0.01

DEFAULT_TRANSPORT = 'File'
DEFAULT_FORMATTER = 'Custom'

lgr = logger.init()


def _set_global_verbosity_level(is_verbose_output=False):
    """sets the global verbosity level for console and the lgr logger.

    :param bool is_verbose_output: should be output be verbose
    """
    global verbose_output
    # TODO: (IMPRV) only raise exceptions in verbose mode
    verbose_output = is_verbose_output
    if verbose_output:
        lgr.setLevel(logging.DEBUG)
    else:
        lgr.setLevel(logging.INFO)
    # print 'level is: ' + str(lgr.getEffectiveLevel())


def _import_config(config_file):
    """returns a Feedr configuration object

    :param string config_file: path to config file
    """
    # get config file path
    config_file = config_file or os.path.join(os.getcwd(), DEFAULT_CONFIG_FILE)
    lgr.debug('config file is: {0}'.format(config_file))
    # append to path for importing
    sys.path.append(os.path.dirname(config_file))
    try:
        lgr.debug('importing generator dict...')
        return __import__(os.path.basename(os.path.splitext(
            config_file)[0])).GENERATOR
    # TODO: (IMPRV) remove from path after importing
    except ImportError:
        lgr.warning('config file not found: {0}.'.format(config_file))
        raise FeedrError('missing config file')
    except SyntaxError:
        lgr.error('config file syntax is malformatted. please fix '
                  'any syntax errors you might have and try again.')
        raise FeedrError('bad config file')


def get_current_time():
    """returns the current time (no microseconds tho)"""
    return datetime.datetime.now().replace(microsecond=0)


def calculate_throughput(elapsed_time, messages):
    """calculates throughput and extracts the number of seconds for the
    run from the elapsed time

    :param elapsed_time: run time
    :param int messages: number of messages to write
    :return: throughput and seconds
    :rtype: tuple
    """
    ftr = [3600, 60, 1]
    seconds = sum([a * b for a, b in zip(
        ftr, [int(i) for i in str(elapsed_time).split(":")])])
    try:
        return messages / seconds, seconds
    except ZeroDivisionError:
        lgr.warning('caluclating throughput for less-than-a-second '
                    'runs is currently not supported. come back soon..')
        return 'Unknown', seconds


def send(instance, client, formatter, format_config, messages, gap, batch):
    """sends data and prints the time it took to send it

    :param instance: transport class instance
    :param client: client to use to send send
    :param string format: formatter to use
    :param dict format_config: formatter configuration to use
    :param int messages: number of messages to send
    :param float gap: gap in seconds between 2 messages
    :param int batch: number of messages per batch
    """
    message_count = 0
    lgr.debug('configuring formatter...')
    # get formatter instance
    # TODO: (IMPRV) move formatter instance definition to function inside
    # TODO: (IMPRV) the current function and add _
    if hasattr(forms, formatter):
        formatter_instance = getattr(forms, formatter)(format_config)
    else:
        lgr.error('could not find formatter: {0}. please make sure the '
                  'formatter you\'re calling exists.'.format(formatter))
        raise FeedrError('missing formatter')
    # and get the current time
    start_time = get_current_time()
    lgr.debug('start time is: {0}'.format(start_time))
    lgr.info('transporting data... EN GARDE!')
    while True:
        # generate the data from the formatter
        data = [formatter_instance.generate_data() for i in xrange(batch)]
        # and send the data through the relevant transport
        instance.send(client, data)
        message_count += batch
        # check if the number of messages sent are less than the desired amount
        if message_count < messages or messages == 0:
            # and sleep the desired amount of time.. zzz zz zZZ zZZzzzz
            sleep(gap)
        else:
            break
        # just to get some feedback during execution
        if not message_count % (1 / gap):
            lgr.info('{0} data pieces written. NICE!'.format(message_count))
    # then get the current time once more
    end_time = get_current_time()
    lgr.debug('end time is: {0}'.format(end_time))
    # and the elapsed time
    elapsed_time = end_time - start_time
    # meH!
    throughput, seconds = calculate_throughput(elapsed_time, messages)
    # TODO: (FEAT) add the option to send the throughput as well to benchmark
    # TODO: (FEAT) the transport process itself.
    lgr.info('DONE! (after {0}h ({1} seconds) with '
             'throughput: {2} pieces/sec. now you can go for coffee.)'.format(
                 elapsed_time, seconds, throughput))
    try:
        # create a pretty table to write the statistical data to
        # TODO: (IMPRV) move this to generator function.
        data = instance.get_data()
        lgr.info('statistical data:\n {0}'.format(data))
    except AttributeError:
        lgr.debug(
            'statistical data not implemented for chosen transport.')
    # TODO: (IMPRV) why is this here?
    return


def config_transport(transports, transport, transport_config):
    """returns a configured instance and client for the transport

    :param transports: transport classes to choose from.
    :param string transport: transport to use
    :param dict transport_config: transport configuration
    """
    lgr.debug('configuring transport...')
    # get transport instance
    if hasattr(transports, transport):
        transport_instance = getattr(transports, transport)(transport_config)
    else:
        lgr.error('could not find transport: {0}. please make sure the '
                  'transport you\'re calling exists.'.format(transport))
        raise FeedrError('missing transport: {0}'.format(transport))
    # get transport client
    client = transport_instance.configure()
    return transport_instance, client


def generator(config=None, transport=None,
              formatter=None, gap=None,
              messages=0, batch=1, verbose=False):
    """generates data

    this will generate data in the requested format and protocol.

    :param string config: path to config file path
    :param string transport: transport type to use
    :param string formatter: formatter to use
    :param float gap: gap in seconds between 2 messages
    :param int messages: number of messages to send
    :param int batch: number of messages to stack before sending
    :param bool verbose: sets verbose state for internal logging.
    """
    def return_real(gap, messages, batch):
        return float(gap), abs(int(messages)), int(batch)
    # set verbosity level for internal logging
    _set_global_verbosity_level(verbose)
    # set params for basic Feedr configuration
    # import config file
    transport = transport if transport else DEFAULT_TRANSPORT
    formatter = formatter if formatter else DEFAULT_FORMATTER
    config = _import_config(config) if config else {}
    gap, msgs, batch = return_real(
        gap if gap else DEFAULT_GAP,
        messages if messages else 0,
        batch if batch else 1
    )

    # TODO: (IMPRV) move config to different function.
    # declare transport and formatter configuration. will assume defaults
    # if config file wasn't imported.
    all_transports = config.get('transports', {}) \
        if config else {}
    all_formatters = config.get('formatters', {}) \
        if config else {}
    transport_config = all_transports.get(transport, {}) \
        if all_transports.get(transport) else {}
    formatter_config = all_formatters.get(formatter, {}) \
        if all_formatters.get(formatter) else {}
    transport = transport_config.get('type', transport) \
        if transport_config else transport
    formatter = formatter_config.get('type', formatter) \
        if formatter_config else formatter

    lgr.debug('transport: {0}'.format(transport))
    lgr.debug('formatter: {0}'.format(formatter))
    lgr.debug('gap: {0}'.format(gap))
    lgr.debug('message count: {0}'.format(msgs if msgs > 0 else 'infinite'))

    # well.. you can't have that right? that would be stupid.
    if msgs > 0 and batch > msgs:
        raise FeedrError('batch number larger than total amount of messages')
    else:
        lgr.debug('batch: {0}'.format(batch))
    # define transport class instance and shipping client
    instance, client = config_transport(
        trans, transport + 'Transport', transport_config)
    # send the stuff
    send(instance, client, formatter + 'Formatter', formatter_config,
         msgs, gap, batch)
    # maybe close a connection to the host is required...
    try:
        instance.close()
    except AttributeError:
        lgr.debug('connection closing not implemented for chosen transport.')


def list_fake_types():
    """prints a list of random data types with an example"""
    fake = Factory.create()
    ignore_list = [
        'add_provider',
        'format',
        'get_formatter',
        'set_formatter',
        'parse',
        'provider',
        'providers',
        'get_providers',
    ]
    fakes_list = []

    # yes, the following is sort of disgusting. will have to find a better way
    # to implement this...
    # list from fake-factory
    for fake_type in dir(fake):
        if not fake_type.startswith('_') and fake_type not in ignore_list:
            fakes_list.append('*** {0} ({1})'.format(
                fake_type, getattr(fake, fake_type)()))
    # list from format mappings default handler
    for fake_type, data in fm.DATA.items():
        fakes_list.append('*** {0} ({1})'.format(
            fake_type, random.choice(data)))
    # list from format mappings additional handlers
    for fake_type in dir(fm.InHouseFaker):
        if not fake_type.startswith('_') and not fake_type == 'default':
            fakes_list.append('*** {0} ({1})'.format(
                fake_type, getattr(fm.InHouseFaker(), fake_type)()))
    print("\n".join(fakes_list))


def list_transports():
    for transport in dir(trans):
        if 'Transport' in transport:
            print('*** {0}'.format(transport.replace('Transport', '')))


def list_formatters():
    for formatter in dir(forms):
        if 'Formatter' in formatter:
            print('*** {0}'.format(formatter.replace('Formatter', '')))


class FeedrError(Exception):
    pass



# flake8: NOQA
"""Script to run feedr via command line

Usage:
    mouth feed [--config=<path> --transport=<string> --format=<string> --gap=<int> --messages=<int> --batch=<int> -v]
    mouth list (fake|transports|formatters)
    mouth --version

Arguments:
    feed                Generates Logs
    list fake           Lists different data options (with an example)
    list transports     Lists the type of transports available
    list formatters     Lists the type of formatters available

Options:
    -h --help                   Show this screen.
    -c --config=<path>          Path to generator config file.
    -t --transport=<string>     transport to use (e.g. File)
    -f --format=<string>        output format (e.g. Json)
    -g --gap=<int>              Number of seconds between messages (can be less than 1)
    -m --messages=<int>         Number of messages to generate (if omitted will assume infinity (good for daemonization))
    -b --batch=<int>            Number of messages to stack before shipping
    -v --verbose                a LOT of output
    --version                   Display current version of feedr and exit

"""

from __future__ import absolute_import
from docopt import docopt
import feedr.logger as logger
from feedr.feedr import _set_global_verbosity_level
from feedr.feedr import generator
from feedr.feedr import list_fake_types, list_transports, list_formatters

lgr = logger.init()


def ver_check():
    import pkg_resources
    version = None
    try:
        version = pkg_resources.get_distribution('feedr').version
    except Exception as e:
        print(e)
    finally:
        del pkg_resources
    return version


def feeder_run(o):
    if o['list']:
        if o['fake']:
            list_fake_types()
        elif o['transports']:
            list_transports()
        elif o['formatters']:
            list_formatters()

    elif o['feed']:
        generator(config=o.get('--config'),
                 transport=o.get('--transport'),
                 formatter=o.get('--format'),
                 gap=o.get('--gap'),
                 messages=o.get('--messages'),
                 batch=o.get('--batch'),
                 verbose=o.get('--verbose'))


def feedr(test_options=None):
    """Main entry point for script."""
    version = ver_check()
    options = test_options or docopt(__doc__, version=version)
    _set_global_verbosity_level(options.get('--verbose'))
    lgr.debug(options)
    feeder_run(options)


def main():
    feedr()


if __name__ == '__main__':
    main()



# flake8: NOQA
from time import gmtime, strftime
import datetime
import random


DATA = {
    'syslog_error_levels': ['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'],
    'syslog_error_levels_lower': ['debug', 'info', 'warning', 'error', 'critical'],
    'http_verbs': ['OPTIONS', 'GET', 'HEAD', 'POST', 'PUT', 'DELETE'],
    'http_versions': ['HTTP/1.0', 'HTTP/1.1'],
    'http_error_codes': [
        '100', '101',
        '200', '201', '202', '203', '204', '205', '206',
        '300', '301', '302', '303', '304', '305',
        '400', '401', '402', '403', '404', '405', '406', '407', '408', '409', '410', '411', '412', '413', '414', '415',
        '500', '501', '502', '503', '504', '505'
    ],
    'http_more_verbs': [
        'TRACE', 'CONNECT', 'PROPFIND', 'PROPPATCH', 'MKCOL',
        'COPY', 'MOVE', 'LOCK', 'UNLOCK', 'VERSION-CONTROL',
        'REPORT', 'CHECKOUT', 'CHECKIN', 'UNCHECKOUT', 'MKWORKSPACE',
        'UPDATE', 'LABEL', 'MERGE', 'BASELINE-CONTROL', 'MKACTIVITY',
        'ORDERPATCH', 'ACL', 'draft-dusseault-http-patch', 'PATCH',
        'draft-reschke-webdav-search', 'SEARCH'
    ],
    'time_zone_number': [
        '-1200', '-1100', '-1000', '-0900', '-0800', '-0700', '-0600', '-0500', '-0400', '-0300', '-0200', '-0100', '-0000',
        '+1200', '+1100', '+1000', '+0900', '+0800', '+0700', '+0600', '+0500', '+0400', '+0300', '+0200', '+0100', '+0000',
    ],
    'month_name_short': ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec'],
    'day_of_week_short': ['Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat'],
}


class InHouseFaker():
    def __init__(self):
        self.dt = datetime.datetime.now()

    def default(self, data_type):
        return random.choice(DATA[data_type])

    def current_time(self):
        return str(self.dt.time().replace(microsecond=0))

    def current_date_time(self):
        return str(self.dt.replace(microsecond=0))

    def current_day_of_month(self):
        return str(self.dt.strftime("%d"))

    def current_day_of_week_short(self):
        return str(self.dt.strftime("%A")[0:3])

    def current_year(self):
        return str(self.dt.strftime("%y"))

    def current_month_name_short(self):
        return str(self.dt.strftime("%B")[0:3])

    def current_time_zone_number(self):
        return strftime("%z", gmtime())
